#=======================================================================
#	$Id: Triples.pm,v 1.2 2005/09/07 10:00:45 pythontech Exp $
#	Virtual base class for storage of triples
#	Copyright (C) 2004,2005  Python Technology Limited
#
#	This program is free software; you can redistribute it and/or
#	modify it under the terms of the GNU General Public License
#	as published by the Free Software Foundation; either version 2
#	of the License, or (at your option) any later version.
#
#	This program is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with this program; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  
#	02111-1307, USA.
#=======================================================================
package PythonTech::Triples;
use strict;

sub new {
    my $class = shift;
    die "Called virtual ${class}::new\n";
}

sub configure {
    my $self = shift;
    die "Called virtual ${self}->configure\n";
}

#-----------------------------------------------------------------------
#	Copy statements to another triple store
#-----------------------------------------------------------------------
sub copy {
    my($self, $other) = @_;
    foreach my $triple ($self->query(undef,undef,undef)) {
	$other->add(@$triple);
    }
}

#-----------------------------------------------------------------------
#	Add a statement in the database
#-----------------------------------------------------------------------
sub add {
    my($self, $subj,$pred,$obj) = @_;
    die "Called virtual ${self}->add\n";
}

#-----------------------------------------------------------------------
#	Remove a statement from the database
#-----------------------------------------------------------------------
sub remove {
    my($self, $subj,$pred,$obj) = @_;
    die "Called virtual ${self}->remove\n";
}

#-----------------------------------------------------------------------
#	Test a statement
#-----------------------------------------------------------------------
sub test {
    my($self, $subj,$pred,$obj) = @_;
    return $self->query($subj,$pred,$obj);
}

#-----------------------------------------------------------------------
#	Run a general query
#-----------------------------------------------------------------------
sub query {
    my($self, $subj,$pred,$obj) = @_;
    die "Called virtual ${self}->query\n";
}

#-----------------------------------------------------------------------
#	Given subject and predicate, list objects
#-----------------------------------------------------------------------
sub sp2o {
    my($self, $subj,$pred) = @_;
    my @res = $self->query($subj,$pred,undef);
    if (wantarray) {
	return map {$_->[0]} @res;
    } else {
	return @res && $res[0]->[0];
    }
}

#-----------------------------------------------------------------------
#	Given predicate and object, list subjects
#-----------------------------------------------------------------------
sub po2s {
    my($self, $pred,$obj) = @_;
    my @res = $self->query(undef,$pred,$obj);
    if (wantarray) {
	return map {$_->[0]} @res;
    } else {
	return @res ? $res[0]->[0] : undef;
    }
}

#-----------------------------------------------------------------------
#	Given subject, list predicates
#-----------------------------------------------------------------------
sub s2p {
    my($self, $subj) = @_;
    my @res = $self->query($subj,undef,undef);
    if (wantarray) {
	my %p;
	foreach (@res) { $p{$_->[0]} = 1 };
	return keys %p;
    } else {
	return @res ? $res[0]->[0] : undef;
    }
}

#-----------------------------------------------------------------------
#	Commit or back out of pending changes
#-----------------------------------------------------------------------
sub commit {
    die ref($_[0])." does not implement 'commit'\n";
}
sub rollback {
    die ref($_[0])." does not implement 'rollback'\n";
}

#-----------------------------------------------------------------------
#	Disconnect from / reconnect to persistence layer
#-----------------------------------------------------------------------
sub disconnect {
    die ref($_[0])." does not implement 'disconnect'\n";
}
sub reconnect {
    die ref($_[0])." does not implement 'reconnect'\n";
}

#-----------------------------------------------------------------------
#	Write as N-Triples
#-----------------------------------------------------------------------
sub write_ntriples {
    my($self, $filename) = @_;
    if (! defined $filename) {
	open(NT,'>&STDOUT') || die "Cannot open stdout: $!\n";
    } else {
	open(NT,'>',$filename) || die "Cannot open $filename: $!\n";
    }
    foreach my $triple ($self->query(undef,undef,undef)) {
	my($subj,$pred,$obj) = @$triple;
	my $tobj;
	if ($obj =~ /^=/o) {
	    # Escape backslash and quotes with backslash...
	    $obj =~ s/[\\\"]/\\$&/g;
	    # Newline as \n, return as \r, tab as \t
	    $obj =~ s/\n/\\n/g;
	    $obj =~ s/\r/\\r/g;
	    $obj =~ s/\t/\\t/g;
	    # Non-ASCII as \uXXXX
	    $obj =~ s/[\x00-\x1f\x7f]/sprintf("\\u%04X",ord($&))/eg;
	    $tobj = '"' . substr($obj,1) . '"';
	} else {
	    $tobj = "<$obj>";
	}
	print NT "<$subj> <$pred> $tobj .\n";
    }
    close(NT);
}

#-----------------------------------------------------------------------
#	Parse a line containing an N-Triple
#-----------------------------------------------------------------------
sub parse_ntriple_line {
    my($self,$line) = @_;
    if ($line =~ /\S/) {
	if (my($subj,$pred,$ouri,$olit) =
	    $line =~ /^\s*
	    <([^\n>]+)>		# Subject as URI
	    \s+
	    <([^\n>]+)>		# Subject as URI
	    \s+
	    (?:<([^\n>]*)>	# Object as URI
	     |
	     \"((?:\\.|[^\\\n\"])*)\" # Object as escaped literal
	     (?:@\w+)?		# Ignore language tag
	     )
	    \s+
	    \.\s*\n?$/x) {
#	    if (0== (($subj) = $line =~ /^\s*<([^\n>]*)>/g)) {
#		die "Bad subject at line $lineno\n";
#	    } elsif (0== (($pred) = $line =~ /\G\s+<([^\n>]*)>/gc)) {
#		die "Bad predicate at line $lineno\n";
#	    } elsif (0== (($ourl,$olit) = $line =~ /\G\s+
#		die "Bad object at line $lineno\n";
#	    } else {
	    if (defined($ouri)) {
		$self->add($subj,$pred,$ouri);
	    } else {
		$olit =~ s/\\u(....)/chr($1)/eg;
		$olit =~ s/\\t/\t/g;
		$olit =~ s/\\r/\r/g;
		$olit =~ s/\\n/\n/g;
		$olit =~ s/\\(.)/$1/g;
		$self->add($subj,$pred,"=$olit");
	    }
	} else {
	    die "N-Triple unrecognised: $line\n";
	}
    }
}

#-----------------------------------------------------------------------
#	Read ntriples from a file handle
#-----------------------------------------------------------------------
sub read_ntriples {
    my($self,$fh) = @_;
    my $lineno = 0;
    while (defined(my $line = <$fh>)) {
	++ $lineno;
	if (! eval {$self->parse_ntriple_line($line), 1}) {
	    chomp(my $err = @_);
	    die "$err at line $.\n";
	}
    }
}

1;
